/*
 * hst_opencl_kernels.c
 * This file is part of PyHST
 *
 * Copyright (C) 2011 - Matthias Vogelgesang
 * Copyright (C) 2002-2008 the European Synchrotron Radiation Facility (ESRF).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PyHST is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

__kernel void hst_filter(__global float *data, __global float *filter)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int id = idy*get_global_size(0) + idx;
    data[id] = data[id] * filter[idx];
}

__kernel void hst_normalize_idft(__global float *data)
{
    const int idx = get_global_id(0);
    const int dim_fft = get_global_size(0);
    data[2*idx] = data[2*idx] / dim_fft;
}

__kernel void hst_spread(__global float *out, __global float *in, const int num_bins)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int dpitch = get_global_size(0)*2;

    /* May diverge but not possible to reduce latency, because num_bins can 
       be arbitrary and not be aligned. */
    if (idx >= num_bins) {
        out[idy*dpitch + idx*2] = 0.0;
        out[idy*dpitch + idx*2 + 1] = 0.0;
    }
    else {
        out[idy*dpitch + idx*2] = in[idy*num_bins + idx];
        out[idy*dpitch + idx*2 + 1] = 0.0;
    }
}

__kernel void hst_pack(__global float *in, __global float *out, const int num_bins)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int dpitch = get_global_size(0)*2;

    if (idx < num_bins) {
        out[idy*num_bins + idx] = in[idy*dpitch + idx*2];
    }
}

__kernel void hst_backproject(const int num_proj,
    const int num_bins,
    const float off_x,
    const float off_y,
    __constant float *cos_table,
    __constant float *sin_table,
    __constant float *axis_table,
    __global float *sinogram,
    __global float *slice)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int slice_width = get_global_size(0);

    /* TODO: maybe use float4 for optimal vectorization on Intel CPUs */
    float2 b = (float2) (idx + off_x, idy + off_y);
    float corr;
    float sum = 0.0;

#pragma unroll 8
    for(int proj = 0; proj < num_proj; proj++) {
        float2 s = (float2) (cos_table[proj], sin_table[proj]);
        s = s*b;
        corr = axis_table[proj];
        sum += sinogram[(int)(proj*num_bins + corr + s.x - s.y)];
    }
    slice[idy*slice_width + idx] = sum;
}
